/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#include <oskit/x86/multiboot.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/pc/base_multiboot.h>
#include <oskit/lmm.h>

extern char **environ;

static char prog_name[] = "kernel";
static char *null_args[2] = {prog_name, 0};

static const char delim[] = " \f\n\r\t\v";

void base_multiboot_init_cmdline(int *argcp, char ***argvp)
{
	int argc;
	char **argv;

	if (boot_info.flags & MULTIBOOT_CMDLINE)
	{
		char *cl = (char*)phystokv(boot_info.cmdline);
		unsigned cllen = strlen(cl);
		char *targ[1 + cllen], *tvar[cllen];
		unsigned narg = 0, nvar = 0;
		char *tok;

		/* Supply the program name as the first argument.  */
		targ[narg++] = prog_name;

		/* Parse out the tokens in the command line.
		   XXX Might be good to handle quotes.  */
		for (tok = strtok(cl, delim); tok; tok = strtok(0, delim))
		{
			if (strchr(tok, '='))
				tvar[nvar++] = tok;
			else
				targ[narg++] = tok;
		}

		/* Copy the pointer arrays to permanent heap memory.  */
		argv = lmm_alloc(&malloc_lmm,
				 sizeof(char*) * (narg + nvar + 2), 0);
		if (!argv) panic("No memory to parse command line");

		memcpy(argv, targ, sizeof(char*) * narg);
		argv[narg] = 0;
		argc = narg;

		environ = &argv[narg+1];
		memcpy(environ, tvar, sizeof(char*) * nvar);
		environ[nvar] = 0;
	}
	else
	{
		/* No command line.  */
		argc = 1;
		argv = null_args;
		environ = null_args + 1;
	}

	*argcp = argc;
	*argvp = argv;
}

