/*
 * Copyright (c) 1997-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */
/*
 * Allocate a file descriptor
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fd.h>

/* start out with NDFILE */
#define NDFILE          20 
#define NDEXTENT        50

/* then extend by `extendby' - which is doubled every time */
static	int extendby; 

int	
fd_alloc(oskit_iunknown_t *obj, int min_fd)
{
	int		i;
	oskit_size_t 	bytes;
	fd_t		*newarray;

try_again:
	/* search for a free fd */
	for (i = min_fd; i < fd_arraylen; i++) {
		if (fd_array[i].obj == 0) {
			/* found an unused slot */
			fd_array[i].obj = obj;
			oskit_iunknown_addref(obj);
			return i;
		}
	}

	/* create the fd_array if we don't have one yet */
	if (fd_arraylen == 0) {
		bytes = NDFILE * sizeof(fd_t);
		fd_array = (fd_t *)malloc(bytes);
		if (fd_array == NULL) {
			errno = ENOMEM;
			return -1;
		}
		memset(fd_array, 0, bytes);
		fd_arraylen = NDFILE;
		extendby = NDEXTENT;
		goto try_again;
	}

	/* otherwise try to extend the existing fd_array */
	bytes = (fd_arraylen + extendby) * sizeof(fd_t);
	newarray = (fd_t *)realloc(fd_array, bytes);
	if (newarray == NULL) {
		errno = ENOMEM;
		return -1;
	}
	fd_array = newarray;

	/* zero out new part of the array */
	memset(fd_array + fd_arraylen, 0, extendby * sizeof(fd_t));
	fd_arraylen += extendby;
	extendby *= 2;
	goto try_again;
}

