/*
 * Copyright (c) 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <threads/pthread_internal.h>

/*
 * Wake up all threads waiting on a condition.
 */
int
pthread_cond_broadcast(pthread_cond_t *c)
{
	int			s, resched = 0;
	pthread_thread_t	*pnext;
	
	s = splhigh();
	pthread_lock(&(c->lock));

	if (queue_empty(&(c->waiters))) {
		pthread_unlock(&(c->lock));
		splx(s);
		return 0;
	}

	queue_iterate(&(c->waiters), pnext, pthread_thread_t *, chain) {
		if (pnext == CURPTHREAD())
			panic("pthread_cond_broadcast: Bad queue!");

		/*
		 * Note race condition. Someone can try to cancel this
		 * thread between the time it comes off the queue, but
		 * before the state gets changed. pthread_cancel will
		 * have to deal with this.
		 */
		pthread_lock(&pnext->lock);
		pnext->flags &= ~THREAD_CONDWAIT;
		pnext->waitcond = 0;
		pthread_unlock(&pnext->lock);
		if (pthread_sched_setrunnable(pnext))
			resched++;
	}
	queue_init(&(c->waiters));
	pthread_unlock(&(c->lock));
	splx(s);

	/*
	 * Yield if no longer highest priority.
	 */
	if (resched)
		pthread_yield();

        return 0;
}
