/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Exit code.
 */
#include <threads/pthread_internal.h>

/*
 * Just a dummy lock to satisfy the switch code.
 */
pthread_lock_t		threads_exit_lock    = PTHREAD_LOCK_INITIALIZER;

void
pthread_exit(void *status)
{
	pthread_thread_t  *pthread = CURPTHREAD();
	int		  p;

	p = splhigh();
	pthread_lock(&pthread->lock);

	pthread_exit_locked(status);

	/*
	 * Never returns.
	 */
}

/*
 * Internal version. Thread lock is locked.
 */
void
pthread_exit_locked(void *status)
{
	pthread_thread_t  *pthread = CURPTHREAD();
		
	if (threads_debug)
		printf("pthread_exit(%d): 0x%x(%d) 0x%x\n",
		       THISCPU, (int) pthread, pthread->tid, (int) status);

	if (pthread->flags & THREAD_EXITING) {
		panic("pthread_exit: Recursive call: 0x%x(%d)\n",
		      (int) pthread, pthread->tid);
	}

	pthread->flags  |= THREAD_EXITING;
	pthread->exitval = (oskit_u32_t) status;
	pthread_unlock(&pthread->lock);

	/*
	 * A little iffy. Drop the spl and do whatever callouts are
	 * required. No locks are held!
	 */
	spllow();

	/*
	 * Call the cancelation handlers.
	 */
	pthread_call_cleanup_handlers();
		
	/*
	 * Call the key destructors.
	 */
	pthread_call_key_destructors();

#ifdef  CPU_INHERIT
	pthread_sched_exit();
#else
#ifdef  PRI_INHERIT
	/*
	 * If this thread is inheriting from a thread (which means a thread
	 * or threads is waiting on it), its probably a bad thing cause
	 * some resource is locked up.
	 */
	if (! queue_empty(&pthread->waiters))
		printf("PTHREAD_EXIT: TID(%d) is exiting, but other "
		       "threads are waiting for it. DEADLOCK?\n",
		       pthread->tid);
#endif
	/*
	 * Stick this thread on the reaperq and fire up the reaper thread.
	 */
	pthread_mutex_lock(&threads_reaperq_mutex);
	queue_enter(&threads_reaperq, pthread, pthread_thread_t *, chain);
	pthread_cond_signal(&threads_reaperq_cond);
	pthread_mutex_unlock(&threads_reaperq_mutex);

	/*
	 * and block ...
	 */
	splhigh();
	pthread_lock(&pthread->lock);
	pthread_sched_reschedule(RESCHED_BLOCK, &pthread->lock);
#endif
	/*
	 * Never returns.
	 */
}

/*
 * Called for an interrupted thread (thread in osenv_sleep). Instead of
 * killing the thread immediately, its allowed to return with an EINTR.
 * The death is caught later, calling this reaper function. The same stuff
 * needs to be done, but the exit status is undefined.
 */
void
pthread_reap(void)
{
	pthread_exit((void *) PTHREAD_CANCELED);

	/*
	 * Never returns.
	 */
}

